return (function ()
  local mesh=love.graphics.newMesh({{0,0,0,0},{1,0,1,0},{1,1,1,1},{0,1,0,1}})
  local effect=love.graphics.newShader [[
    extern vec2 offset;
    extern vec2 size;
    extern vec2 mapsize;
    extern sampler2D tiles;
    extern sampler2D palette;
    vec4 effect(vec4 color,Image texture,vec2 texture_coords,vec2 pixel_coords)
    {
      texture_coords*=size;
      texture_coords+=offset;
      vec4 cell=Texel(texture,floor(texture_coords)/mapsize);
      vec2 pixelpos=mod(texture_coords,1);
      if(cell.a>=.5)pixelpos.x=1-pixelpos.x;
      if(mod(cell.a,.5)>=.25)pixelpos.y=1-pixelpos.y;
      if(mod(cell.a,.25)>=.125)pixelpos=pixelpos.yx;
      return Texel(palette,vec2(cell.b,Texel(tiles,pixelpos/16+cell.rg).r));
    }
  ]]
  local make=function (width,height)
    local cells={}
    for y=0,height-1 do
      cells[y]={}
      for x=0,width-1 do
        cells[y][x]={n=0,pal=0,xflip=false,yflip=false,dflip=false}
      end
    end
    local data=love.image.newImageData(width,height)
    local image=love.graphics.newImage(data)
    image:setWrap("repeat","repeat")
    image:setFilter("nearest","nearest")
    return {setCell=function(x,y,n,pal,xflip,yflip,dflip)
      cells[y][x]={n=n,pal=pal,xflip=xflip,yflip=yflip,dflip=dflip}
    end, flip=function()
      local ptr=ffi.cast('unsigned char*',data:getPointer())
      for y=0,height-1 do
        for x=0,width-1 do
          local c=cells[y][x]
          ptr[0]=(c.n%16)*16
          ptr=ptr+1
          ptr[0]=math.floor(c.n/16)*16
          ptr=ptr+1
          ptr[0]=c.pal*8
          ptr=ptr+1
          ptr[0]=(c.xflip and 128 or 0)+(c.yflip and 64 or 0)+(c.dflip and 32 or 0)+1
          ptr=ptr+1
        end
      end
      image:refresh()
    end, draw=function(tiles,palette,x,y,width,height,offset,size)
      love.graphics.setShader(effect)
      mesh:setTexture(image)
      effect:send("tiles",tiles)
      effect:send("palette",palette)
      effect:send("offset",offset)
      effect:send("size",size)
      effect:send("mapsize",{image:getWidth(),image:getHeight()})
      love.graphics.draw(mesh,x,y,0,width,height)
      love.graphics.setShader()      
    end}
  end
  return {new=make}
end)()
